#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>

#include <opencv2/opencv.hpp>

const char SERIAL_PORT[] = "/dev/ttyUSB0";

int main(int argc,char** argv)
{
    struct termios tio;
    int tty_fd;
    short int s;

    memset(&tio, 0, sizeof(tio));
    tio.c_iflag = 0;
    tio.c_oflag = 0;
    tio.c_cflag = CS8 |CREAD | CLOCAL; // 8n1, see termios.h for more information
    tio.c_lflag = 0;
    tio.c_cc[VMIN] = 1;
    tio.c_cc[VTIME] = 5;

    tty_fd = open(SERIAL_PORT, O_RDWR);// | O_NONBLOCK);

    if(tty_fd < 0) {
        printf("Need to be root\n");
        return 0;
    }

    cfsetospeed(&tio, B1500000);
    cfsetispeed(&tio, B1500000);

    tcsetattr(tty_fd, TCSANOW, &tio);

    cv::Mat grey = cv::imread("Lenna128.png", 0);
    assert(grey.data);
    cv::imshow("grey", grey);

    cv::Mat filtered = cv::Mat::zeros(128-2, 128-2, CV_16S);

    int64_t start = cv::getTickCount();

    // Bootstrap initial two rows
    write(tty_fd, grey.data, grey.cols*2);

    for(int y=2; y < grey.rows; y++) {
        printf("%d\n", y);

        write(tty_fd, grey.ptr(y), grey.cols);

        // read pixels
        short *dst = filtered.ptr<short>(y-2);

        for(int x=0; x < grey.cols-2; x++) {
            unsigned char a, b;

            read(tty_fd, &a, 1);
            read(tty_fd, &b, 1);

            s = a + (b << 8);

            dst[x] = s;
        }
    }

    // display
    double min_val, max_val;
    cv::minMaxLoc(filtered, &min_val, &max_val);

    cv::Mat img = (filtered - min_val)*255/(max_val - min_val);
    img.convertTo(img, CV_8U);

    cv::imshow("filtered", img);
    cv::waitKey(1);

    printf("Time: %f ms\n", (cv::getTickCount() - start)*1000.0 / cv::getTickFrequency());

    cv::waitKey(0);

    close(tty_fd);

    return 0;
}
