`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    00:38:00 02/21/2015 
// Design Name: 
// Module Name:    serial 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module Serial(
	input wire reset,
	input wire clk,
	input wire rx,
	output reg tx,
	input wire [7:0] tx_data,
	input wire tx_data_ready,
	output reg tx_busy,
	output reg [7:0] rx_data,
	output reg rx_data_ready
);

parameter CLK_SPEED = 100000000; //50000000;
parameter BAUD_RATE = 1500000;
parameter BAUD_CLK_COUNT = CLK_SPEED / BAUD_RATE;
parameter RECV_SAMPLING = 5; // when recv'ing must sample higher

reg [12:0] count = 0; // for transmitting
reg [12:0] count2 = 0; // for receiving
reg baud_tick = 0;
reg baud_tick_fast = 0;
	 	 
// tx
reg [5:0] tx_state;
		 
// rx
reg [1:0] rx_sample;
reg [5:0] bit_count;
reg rx_bit;
reg [5:0] rx_state;
reg next_bit;
		 
initial begin
	rx_state <= 0;
	tx_state <= 0;
	tx_busy <= 0;
	rx_data_ready <= 0;
	next_bit <= 0;
	bit_count <= 0;
end

always @ (posedge clk)
begin
	count <= count + 1;

	if(count == BAUD_CLK_COUNT)			
	begin
		baud_tick <= 1;
		count <= 0;
	end
	else
		baud_tick <= 0;	
end

always @ (posedge clk)
begin	
	count2 <= count2 + 1;

	if(count2 == BAUD_CLK_COUNT/RECV_SAMPLING)			
	begin
		baud_tick_fast <= 1;
		count2 <= 0;
	end
	else
		baud_tick_fast <= 0;	
end
	
always @ (posedge clk)
begin
	if(reset) tx_state <= 0;
	
	begin
		case(tx_state)
			0: if(baud_tick && tx_data_ready) tx_state <= 1; // idle
			1: if(baud_tick) tx_state <= tx_state + 1;// start bit
			2: if(baud_tick) tx_state <= tx_state + 1; // 0
			3: if(baud_tick) tx_state <= tx_state + 1; // 1
			4: if(baud_tick) tx_state <= tx_state + 1; // 2
			5: if(baud_tick) tx_state <= tx_state + 1; // 3
			6: if(baud_tick) tx_state <= tx_state + 1; // 4
			7: if(baud_tick) tx_state <= tx_state + 1; // 5
			8: if(baud_tick) tx_state <= tx_state + 1; // 6
			9: if(baud_tick) tx_state <= 0; // 7 		
			
			default: tx_state <= 0;
		endcase
	end
end

// tx state machine output
reg last_bit_sent = 0;

always @ (posedge clk)
begin
	if(reset) 
	begin
		tx <= 1;
		tx_busy <= 0;
	end

	case(tx_state)	
		0: begin tx <= 1; tx_busy <= 0; end // idle/stop bit
		1: begin tx <= 0; tx_busy <= 1; end// start bit
		2: tx <= tx_data[0]; 
		3: tx <= tx_data[1]; 
		4: tx <= tx_data[2]; 
		5: tx <= tx_data[3]; 
		6: tx <= tx_data[4]; 
		7: tx <= tx_data[5]; 
		8: tx <= tx_data[6]; 
		9: tx <= tx_data[7]; 
		
	endcase
end

// read the bit, with filtering
always @ (posedge clk)
if(baud_tick_fast)
begin
	rx_sample <= {rx, rx_sample[1]};

	if(rx_sample == 2'b00)
		rx_bit <= 0;
	else if(rx_sample ==2'b11)
		rx_bit <= 1;
end

// generate signal for state machine, only when start bit has been received
always @ (posedge clk)
begin
	if(rx_state == 0)
		bit_count <= 0;
	else if(baud_tick_fast)
	begin
		bit_count <= bit_count + 1;	
		
		if(bit_count == RECV_SAMPLING-1)
		begin
			next_bit <= 1;
			bit_count <= 0;
		end
		else
			next_bit <= 0;		
	end
end

// reciever state machine
always @ (posedge clk)
begin
	if(reset)
	begin
		rx_state <= 0;
		rx_data_ready <= 0;
	end
	
	if(rx_state == 0)
		rx_data_ready <= 0;
		
	if(baud_tick_fast)
	begin	
		case(rx_state)
			0: 
				begin 			
					rx_data_ready <= 0;
					if(~rx_bit) // start bit
						rx_state <= 1;
				end
			1: if(next_bit) begin rx_data[0] <= rx_bit; rx_state <= 2; end
			2: if(next_bit) begin rx_data[1] <= rx_bit; rx_state <= 3; end
			3: if(next_bit) begin rx_data[2] <= rx_bit; rx_state <= 4; end
			4: if(next_bit) begin rx_data[3] <= rx_bit; rx_state <= 5; end
			5: if(next_bit) begin rx_data[4] <= rx_bit; rx_state <= 6; end
			6: if(next_bit) begin rx_data[5] <= rx_bit; rx_state <= 7; end
			7: if(next_bit) begin rx_data[6] <= rx_bit; rx_state <= 8; end
			8: if(next_bit) begin rx_data[7] <= rx_bit; rx_state <= 9; end	
			9: if(next_bit && rx_bit) // stop bit 
				begin
					rx_state <= 0; 
					rx_data_ready <= 1;		
				end
			default: rx_state <= 0;
		endcase
	end
end

endmodule
