`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    23:33:51 03/30/2015 
// Design Name: 
// Module Name:    top 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module top(
input wire clk,
input wire reset,
input wire rx,
output wire tx,
output reg LED0,
output reg LED1,
output reg LED2,
output reg LED3,
output reg LED4,
output reg LED5,
output reg LED6,
output reg LED7
    );

parameter WIDTH = 128;

reg [7:0] in0;
reg [7:0] in1;
reg [7:0] in2;
reg [7:0] in3;
reg [7:0] in4;
reg [7:0] in5;
reg [7:0] in6;
reg [7:0] in7;
reg [7:0] in8;
wire [15:0] q;

filter3x3 filter(
.clk(clk),
.in0(in0),
.in1(in1),
.in2(in2),
.in3(in3),
.in4(in4),
.in5(in5),
.in6(in6),
.in7(in7),
.in8(in8),
.q(q)
    );

wire tx_busy;
wire [7:0] rx_data;
reg [7:0] tx_data;
wire rx_data_ready;
reg tx_data_ready;

Serial serial(
	.reset(reset),
	.clk(clk),
	.rx(rx),
	.tx(tx),
	.tx_data(tx_data),
	.tx_data_ready(tx_data_ready),
	.tx_busy(tx_busy),
	.rx_data(rx_data),
	.rx_data_ready(rx_data_ready)
);

reg [9:0] addra;
reg [9:0] addrb;

reg wea[0:2] = {0,0,0};
reg web[0:2] = {0,0,0};

reg [7:0] dina = 0;
wire [7:0] douta[0:2];
wire [7:0] doutb[0:2];

blk_mem_gen_v7_3 RAM0(
  .clka(clk),
  .wea(wea[0]),
  .addra(addra),
  .dina(dina),
  .douta(douta[0]),
  .clkb(clk),
  .web(web[0]),
  .addrb(addrb),
  //.dinb(0),
  .doutb(doutb[0])
);

blk_mem_gen_v7_3 RAM1(
  .clka(clk),
  .wea(wea[1]),
  .addra(addra),
  .dina(dina),
  .douta(douta[1]),
  .clkb(clk),
  .web(web[1]),
  .addrb(addrb),
  //.dinb(0),
  .doutb(doutb[1])
);

blk_mem_gen_v7_3 RAM2(
  .clka(clk),
  .wea(wea[2]),
  .addra(addra),
  .dina(dina),
  .douta(douta[2]),
  .clkb(clk),
  .web(web[2]),
  .addrb(addrb),
  //.dinb(0),
  .doutb(doutb[2])
);

///////////////////////////
reg [5:0] state = 0;
reg [10:0] count = 0;

reg [1:0] top = 0;
reg [1:0] mid = 1;
reg [1:0] bottom = 2;

always @ (posedge clk)
begin
	if(reset)
	begin
		count <= 0;
		state <= 0;	
		tx_data_ready <= 0;
		wea[0] <= 0;
		web[0] <= 0;
		wea[1] <= 0;
		web[1] <= 0;
		wea[2] <= 0;
		web[2] <= 0;
		
		LED0 <= 0;
		LED1 <= 0;
		LED2 <= 0;
		LED3 <= 0;
		LED4 <= 0;
		LED5 <= 0;
		LED6 <= 0;
		LED7 <= 0;	
	end
	
	if(state == 0)
		LED0 <= 1;
	if(state == 1)
		LED1 <= 1;		
	else if(state == 2)
		LED2 <= 1;
	else if(state == 3)
		LED3 <= 1;
	else if(state == 4)
		LED4 <= 1;
		
	case(state)	
		0: // bootstrap initial two rows
		if(rx_data_ready)
		begin
			addra <= count;
			dina <= rx_data;
			wea[0] <= 1;
			count <= count + 1;
			
			if(count == WIDTH-1) 
			begin
				count <= 0;
				wea[0] <= 0;
				state <= 1;
			end
		end
		else
			wea[0] <= 0;

		1: // bootstrap initial two rows
		if(rx_data_ready)
		begin
			addra <= count;
			dina <= rx_data;
			wea[1] <= 1;
			count <= count + 1;
			
			if(count == WIDTH-1) 
			begin
				count <= 0;
				wea[1] <= 0;
				state <= 2;
			end
		end
		else
			wea[1] <= 0;
			
		2: // fill the bottom row
		if(rx_data_ready)
		begin
			addra <= count;
			dina <= rx_data;
			wea[bottom] <= 1;
			count <= count + 1;
			
			if(count == WIDTH-1) 
			begin
				count <= 0;
				wea[bottom] <= 0;
				state <= 3;
			end
		end
		else
			wea[bottom] <= 0;		

		// filtering part
		3: // read columns 1, 2
		begin
			addra <= count;
			addrb <= count + 1;
			state <= 4;
		end
		
		4:
		begin		
			in0 <= douta[top];
			in1 <= doutb[top];
			
			in3 <= douta[mid];
			in4 <= doutb[mid];

			in6 <= douta[bottom];
			in7 <= doutb[bottom];
		
			state <= 5;
		end
		
		// read column 3
		5:
		begin
			addra <= count + 2;
			state <= 6;
		end
		
		6:
		begin
			in2 <= douta[top];
			in5 <= douta[mid];
			in8 <= douta[bottom];	

			state <= 7;
		end
			
		// apply 3x3 filter, result on next clock cycle
		7:
			state <= 8;
			
		8: // send first byte
		begin
			tx_data <= q[7:0];
			tx_data_ready <= 1;
			state <= 9;
		end
		
		9: // send started
		if(tx_busy)
			state <= 10;
			
		10: // send done
		if(~tx_busy)
		begin
			tx_data_ready <= 0;
			state <= 11;
		end
		
		11: // send second byte
		begin
			tx_data <= q[15:8];
			tx_data_ready <= 1;			
			state <= 12;
		end		

		12: // send started
		if(tx_busy) 
			state <= 13;
			
		13: // send done
		if(~tx_busy)
		begin
			tx_data_ready <= 0;
			state <= 14;
		end
		
		14:
		if(count == WIDTH-3) // done row
		begin				
			// shuffle the rows
			case(top)
				0:
				begin
					top <= 1;
					mid <= 2;
					bottom <= 0;				
				end
				
				1:
				begin
					top <= 2;
					mid <= 0;
					bottom <= 1;
				end
				
				2:
				begin
					top <= 0;
					mid <= 1;
					bottom <= 2;
				end
			endcase
	
			count <= 0;
			state <= 2;
		end
		else
		begin
			count <= count + 1;
			state <= 3;
		end
		
		default:
			state <= 0;		
	endcase
end

endmodule
